#include <cassert>
#include <cmath>
#include <sys/timeb.h>

#include <algorithm> // 'sort'...
#include <execution> // 'std::execution::par', etc.
#include <initializer_list>
#include <iostream>
#include <random>
#include <stdexcept>
#include <string>
#include <unordered_map>
#include <vector>

#include "Horner.h"

Polynomial::Polynomial(unsigned maximum_degree) {
	std::random_device rd; /* This is used to obtain a seed for the random number engine: http://www.cplusplus.com/reference/random/random_device/operator()... */
	std::mt19937 engine(rd()); /* https://en.cppreference.com/w/cpp/numeric/random/mersenne_twister_engine */
	std::uniform_int_distribution<> draw(0U, maximum_degree);
	std::uniform_real_distribution<> draw_(0.01, 49.90);

	for (unsigned n = 0U; n < maximum_degree; n++)
		// 'draw' transforms the random 'unsigned int' generated by 'engine' into an 'int' belonging to [0, maximum_degree]
		// Keys are unique. So, insertions may lead to no effect:
		_polynomial.insert(std::make_pair<unsigned, double>(draw(engine), draw_(engine)));
	assert(_polynomial.size() <= maximum_degree);
}

Polynomial::Polynomial(const std::initializer_list<std::unordered_map<unsigned, double>::value_type>& polynomial) {
	for (std::unordered_map<unsigned, double>::value_type member : polynomial) // "Range-based 'for'"
		_polynomial.insert(member);
}

double Polynomial::horner_method(double x, bool parallel) const { // Horner method...
	std::vector<long> degrees; // 'std::unordered_map' is unsortable!
	for (const auto& member : _polynomial) // Copy polynomial's degrees before sorting...
		degrees.push_back((long)member.first);

	parallel ?
		// Parallel sorting (C++ 17) based on descending (C++ 14):
		std::sort(std::execution::par, degrees.begin(), degrees.end(), std::greater<unsigned>()) :
		std::sort(std::execution::seq, degrees.begin(), degrees.end(), std::greater<unsigned>());
	// std::sort(std::execution::par_unseq, degrees.begin(), degrees.end(), std::greater<unsigned>()); // Parallel and vectorized

	double coefficient;
	double result = 0;
	for (long n = *degrees.begin(); n > 0L; n--) {
		try {
			coefficient = _polynomial.at((unsigned)n); // '[]' conflicts with 'const' member function
			// std::cout << n << '\t';
		}
		catch (std::out_of_range& oor) {
			coefficient = 0.;
			// std::cerr << n << ": " << oor.what() << '\t';
		}
		result = (result + coefficient) * x;
	}
	try {
		result += _polynomial.at(0U);
	}
	catch (std::out_of_range& oor) {
		// std::cerr << "'x^^0' does not exist: " << oor.what() << '\t';
	}
	return result;
}

double Polynomial::simple_method(double x) const { // "Normal" evaluation...
	double result = 0;
	for (auto& member : _polynomial)
		result += member.second * ::pow(x, (double)member.first);
	return result;
}

std::string Polynomial::to_string() const {
	std::string result;
	for (auto& member : _polynomial)
		result += std::to_string(member.second) + " * x^^" + std::to_string(member.first) + " + ";
	return result.substr(0, result.size() - 3);
}

int main(int argc, char** argv) {
	Polynomial p({ {4U,3.},{2U,2.},{0U,1.} });
	std::cout << p.to_string() << std::endl;
	std::cout << p.horner_method(2.) << std::endl << std::endl;;

	Polynomial p1(1000U); // std::cout << p1.to_string() << std::endl;

	double result;
	timeb start, end;

	::ftime(&start);
	result = p1.horner_method(5.); /* 'false' */
	::ftime(&end);
	std::cout << "start ('p1.horner_method(5., false)'): " << start.time << "." << start.millitm << " - end: " << end.time << "." << end.millitm << "\n";

	::ftime(&start);
	result = p1.horner_method(5., true);
	::ftime(&end);
	std::cout << "start ('p1.horner_method(5., true)'): " << start.time << "." << start.millitm << " - end: " << end.time << "." << end.millitm << "\n";

	return 0;
}

